# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from threading import Thread

import gtk

class PopupMenu(gtk.Menu):
    """ Popup menu """

    def __init__(self, main_instance):
        """ Constructor """
        gtk.Menu.__init__(self)
        self.main = main_instance

        self.toggle_mute = gtk.CheckMenuItem(_("Mute"))
        self.toggle_mute.set_active(self.main.alsactrl.is_muted())
        self.mute_handler_id = self.toggle_mute.connect("toggled", self.on_toggle_mute)
        self.add(self.toggle_mute)

        self.toggle_mixer = gtk.CheckMenuItem(_("Show Mixer"))
        self.toggle_mixer.set_active(self.main.mixer_get_active())
        self.mixer_handler_id = self.toggle_mixer.connect("toggled", self.on_toggle_mixer)
        self.add(self.toggle_mixer)

        item = gtk.ImageMenuItem("gtk-preferences")
        item.connect("activate", self.show_preferences)
        self.add(item)

        item = gtk.ImageMenuItem("gtk-quit")
        item.connect("activate", self.main.quit)
        self.add(item)

        self.show_all()

    def show_preferences(self, widget=None, data=None):
        """ Show preferences window """
        self.main.preferences.open()

    def on_toggle_mute(self, widget=None):
        """ Toggles mute status """
        self.main.alsactrl.set_mute(widget.get_active())
        volume = self.main.get_volume()
        icon = self.main.get_icon_name(volume)
        self.main.update_icon(volume, icon)
        if self.main.show_tooltip:
            self.main.update_tooltip(volume)

    def on_toggle_mixer(self, widget=None):
        """ Start mixer app in new thread """
        Thread(target = self.main.toggle_mixer).start()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import gettext
from subprocess import Popen, PIPE
from signal import SIGTERM

try:
    import gtk
    import gobject
    assert gtk.pygtk_version >= (2, 16, 0)
except ImportError, AssertionError:
    sys.stderr.write("This program needs pygtk 2.16 or higher\nExiting\n")
    sys.exit(1)

try:
    import alsaaudio as alsa
except ImportError:
    sys.stderr.write("This program needs pyalsaaudio 0.6 or higher\nExiting\n")
    sys.exit(1)

try:
    from config import Config
    import preferences
    from preferences import PREFS
    from alsactrl import AlsaControl
    from dbusservice import DBusService
    from scale import VolumeScale
    from menu import PopupMenu
    from utils import which
except ImportError:
    sys.stderr.write("Can't import application modules\nExiting\n")
    sys.exit(1)

config = Config()
gettext.bindtextdomain(config.app_name, config.locale_dir)
gettext.textdomain(config.app_name)

import __builtin__
__builtin__._ = gettext.gettext

class VolumeTray(gtk.StatusIcon):
    """ GTK+ application for controlling audio volume from system tray/notification area """

    def __init__(self):
        """ Constructor """
        gtk.StatusIcon.__init__(self)

        self.config = config
        self.preferences = preferences.Preferences(self)

        self.toggle = PREFS["toggle"]
        self.mixer = PREFS["mixer"]
        self.mixer_internal = PREFS["mixer_internal"]
        self.icon_theme = PREFS["icon_theme"]
        self.show_tooltip = bool(int(PREFS["show_tooltip"]))
        self.run_in_terminal = bool(int(PREFS["run_in_terminal"]))
        self.scale_increment = float(PREFS["scale_increment"])
        self.scale_show_value = bool(int(PREFS["scale_show_value"]))
        self.keys = bool(int(PREFS["keys"]))
        self.keys_backend = PREFS["keys_backend"]
        self.show_notify = bool(int(PREFS["show_notify"]))
        self.notify_timeout = float(PREFS["notify_timeout"])
        self.notify_position = bool(int(PREFS["notify_position"]))
        self.notify_body = PREFS["notify_body"]

        if which("pidof"):
            self.pid_app = "pidof -x"
        elif which("pgrep"):
            self.pid_app = "pgrep"

        try:
            from Xlib import X
            self.has_xlib = True
        except ImportError:
            self.has_xlib = False

        self.lock = False
        self.lockid = None
        self.notify = None
        self.key_press = False
        self.keys_events = None

        self.alsactrl = AlsaControl(PREFS)
        self.menu = PopupMenu(self)
        self.scale = VolumeScale(self)
        self.dbus = DBusService(self)

        if self.keys:
            self.init_keys_events()
        if self.show_notify:
            self.init_notify()

        self.connect("button_press_event", self.on_button_press_event)
        self.connect("scroll_event", self.on_scroll_event)
        self.connect("popup_menu", self.on_popup_menu)

        # set current volume
        self.update()

        # watch for changes
        fd, eventmask = self.alsactrl.get_descriptors()
        self.watchid = gobject.io_add_watch(fd, eventmask, self.update)

    def init_keys_events(self):
        """ Initialize keys events """
        if self.keys_events:
            if hasattr(self.keys_events, "stop"):
                self.keys_events.stop()
            del self.keys_events
            self.keys_events = None

        if not self.keys:
            return

        if self.keys_backend == "hal":
            try:
                from dbusevent import DBusEvent
                self.keys_events = DBusEvent(self)
            except Exception, err:
                sys.stderr.write("%s.%s: %s\n" % (
                    __name__, sys._getframe().f_code.co_name, str(err)))
                self.keys_events = None
        elif self.keys_backend == "xlib":
            if self.has_xlib:
                try:
                    from xlibevent import XlibEvent
                    self.keys_events = XlibEvent(self)
                    self.keys_events.start()
                except Exception, err:
                    sys.stderr.write("%s.%s: %s\n" % (
                        __name__, sys._getframe().f_code.co_name, str(err)))
                    self.keys_events = None
            else:
                sys.stderr.write("Xlib backend needs python-xlib 0.15rc1 or higher\n")
                self.keys_events = None

    def init_notify(self):
        """ Initialize desktop notifications """
        if self.notify:
            self.notify.close()
            del self.notify
            self.notify = None

        if self.show_notify:
            try:
                from notification import Notification
                self.notify = Notification(self)
            except Exception, err:
                sys.stderr.write("%s.%s: %s\n" % (
                    __name__, sys._getframe().f_code.co_name, str(err)))
                self.notify = None

    def on_volume_changed(self, widget=None, data=None):
        """ Callback for scale value_changed signal """
        if self.lock:
            return

        if self.lockid:
            gobject.source_remove(self.lockid)
            self.lockid = None

        self.lock = True
        volume = int(self.scale.slider.get_value())
        self.alsactrl.set_volume(volume)
        vol = self.get_volume()

        icon = self.get_icon_name(vol)
        self.update_icon(vol, icon)
        if self.show_tooltip:
            self.update_tooltip(vol)

        if self.key_press:
            if self.show_notify and self.notify:
                self.update_notify(vol, icon)

        self.lockid = gobject.timeout_add(10, self._unlock)

    def _unlock(self):
        """ Unlock scale """
        self.lock = False
        self.lockid = None
        self.key_press = False
        return False

    def on_button_press_event(self, widget, event, data=None):
        """ Callback for button_press_event """
        if event.button == 1:
            self.scale.toggle_window()
        elif event.button == 2:
            if self.toggle == "mute":
                self.change_volume("mute")
            elif self.toggle == "mixer":
                self.menu.toggle_mixer.set_active(
                        not self.menu.toggle_mixer.get_active())

    def on_scroll_event(self, widget, event):
        """ Callback for scroll_event """
        if event.direction == gtk.gdk.SCROLL_UP:
            self.change_volume("up")
        elif event.direction == gtk.gdk.SCROLL_DOWN:
            self.change_volume("down")
        if self.show_notify and self.notify:
            self.notify.close()

    def on_popup_menu(self, status, button, time):
        """ Show popup menu """
        self.menu.toggle_mixer.handler_block(self.menu.mixer_handler_id)
        self.menu.toggle_mixer.set_active(self.mixer_get_active())
        self.menu.toggle_mixer.handler_unblock(self.menu.mixer_handler_id)

        self.menu.toggle_mute.handler_block(self.menu.mute_handler_id)
        self.menu.toggle_mute.set_active(self.alsactrl.is_muted())
        self.menu.toggle_mute.handler_unblock(self.menu.mute_handler_id)

        self.menu.popup(None, None, gtk.status_icon_position_menu, button, time, self)

    def change_volume(self, event, key_press=False):
        """ Change volume """
        self.key_press = key_press
        volume = self.scale.slider.get_value()

        if event == "up":
            volume = min(100, volume + self.scale_increment)
        elif event == "down":
            volume = max(0, volume - self.scale_increment)

        if event == "mute":
            self.menu.toggle_mute.set_active(
                    not self.menu.toggle_mute.get_active())
        else:
            self.menu.toggle_mute.set_active(False)
            self.set_volume(volume)

    def get_icon_name(self, volume):
        """ Returns icon name for current volume """
        if volume == 0 or volume == _("Muted"):
            icon = "audio-volume-muted"
        elif volume <= 33:
            icon = "audio-volume-low"
        elif volume <= 66:
            icon = "audio-volume-medium"
        elif volume > 66:
            icon = "audio-volume-high"
        return icon

    def get_icon_themes(self):
        themes = ["Default"]
        icons_dir = os.path.join(config.res_dir, "icons")
        try:
            for file in os.listdir(icons_dir):
                if os.path.isdir(os.path.join(icons_dir, file)):
                    if not file.startswith("."):
                        themes.append(file)
        except OSError:
            pass
        return themes

    def get_status_info(self, volume):
        """ Returns status information """
        var = "" if volume == _("Muted") else "%"
        card_name = self.alsactrl.get_card_name()
        mixer_name = self.alsactrl.get_mixer_name()
        return var, card_name, mixer_name

    def set_volume(self, volume):
        """ Set volume """
        if volume != self.scale.slider.get_value():
            self.scale.slider.set_value(volume)
        else:
            self.scale.slider.emit("value_changed")

    def get_volume(self):
        """ Get volume """
        if self.alsactrl.is_muted():
            return _("Muted")
        else:
            return self.alsactrl.get_volume()

    def update_icon(self, volume, icon):
        """ Update icon """
        if self.icon_theme != "Default":
            icon = os.path.abspath(os.path.join(
                    config.res_dir, "icons", self.icon_theme, "32x32", icon+".png"))
            self.set_from_file(icon)
        else:
            self.set_from_icon_name(icon)

    def update_tooltip(self, volume):
        """ Update tooltip """
        var, card_name, mixer_name = self.get_status_info(volume)
        tooltip = "<b>%s: %s%s </b>\n<small>%s: %s\n%s: %s</small>" % (
                _("Output"), volume, var, _("Card"), card_name, _("Mixer"), mixer_name)
        self.set_tooltip_markup(tooltip)

    def update_notify(self, volume, icon):
        """ Update notification """
        if self.icon_theme != "Default":
            icon = os.path.abspath(os.path.join(
                    config.res_dir, "icons", self.icon_theme, "48x48", icon+".png"))
        self.notify.show(icon, self.notify_body, self.notify_timeout, volume)

    def update(self, source=None, condition=None):
        """ Update volume """
        if self.lock:
            gobject.source_remove(self.watchid)
            fd, eventmask = self.alsactrl.get_descriptors()
            self.watchid = gobject.io_add_watch(fd, eventmask, self.update)
            return False
        try:
            self.alsactrl = AlsaControl(preferences.PREFS)
            volume = self.alsactrl.get_volume()
            gtk.gdk.threads_enter()
            self.set_volume(volume)
            gtk.gdk.threads_leave()
            return True
        except Exception, err:
            sys.stderr.write("%s.%s: %s\n" % (
                __name__, sys._getframe().f_code.co_name, str(err)))
            return False

    def toggle_mixer(self, widget=None):
        """ Toggle mixer application """
        mixer = "volti-mixer" if self.mixer_internal else self.mixer
        if not mixer:
            return
        try:
            pid = self.mixer_get_pid()
            if pid:
                os.kill(pid, SIGTERM)
            else:
                if self.run_in_terminal and not self.mixer_internal:
                    term = self.find_term()
                    cmd = [term, "-e", mixer]
                else:
                    cmd = which(mixer)
                Popen(cmd, shell=False)
        except Exception, err:
            sys.stderr.write("%s.%s: %s\n" % (
                __name__, sys._getframe().f_code.co_name, str(err)))

    def find_term(self):
        term = os.getenv("TERM")
        if term == "linux" or term is None:
            if which("gconftool-2"):
                term = Popen(["gconftool-2", "-g",
                    "/desktop/gnome/applications/terminal/exec"],
                        stdout=PIPE).communicate()[0].strip()
            else:
                term = 'xterm'
        if term == "rxvt" and not which(term):
            term = "urxvt"
        return term

    def mixer_get_pid(self):
        """ Get process id of mixer application """
        mixer = "volti-mixer" if self.mixer_internal else self.mixer
        pid = Popen(self.pid_app + " " + os.path.basename(mixer),
                stdout=PIPE, shell=True).communicate()[0]
        if pid:
            try:
                return int(pid)
            except ValueError:
                return None
        return None

    def mixer_get_active(self):
        """ Returns status of mixer application """
        if self.mixer_get_pid():
            return True
        return False

    def main(self):
        """ Main loop """
        gobject.threads_init()
        try:
            gtk.main()
        except KeyboardInterrupt:
            pass

    def quit(self, widget=None):
        """ Quit main loop """
        gtk.main_quit()



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gtk

class VolumeScale():
    """ Volume scale/slider """

    def __init__(self, main_instance):
        """ Constructor """
        self.main = main_instance
        self.screen, self.rectangle, self.orientation = self.main.get_geometry()
        self.win = None
        self.init_window()

    def init_window(self):
        if self.win:
            self.win.unrealize()

        self.win = gtk.Window(type=gtk.WINDOW_POPUP)
        self.win.resize(1, 1)
        frame = gtk.Frame()
        frame.set_shadow_type(gtk.SHADOW_OUT)

        self.init_slider()

        self.align.add(self.slider)
        frame.add(self.align)
        self.win.add(frame)

        self.win.connect("button_press_event", self.on_window_button_press_event)
        self.win.connect("key_release_event", self.on_window_key_release_event)
        self.win.connect("scroll_event", self.on_window_scroll_event)
        self.win.connect_after("realize", self.on_realize)

    def init_slider(self):
        if self.orientation == gtk.ORIENTATION_VERTICAL:
            self.slider = gtk.HScale()
        else:
            self.slider = gtk.VScale()

        self.align = gtk.Alignment()

        self.slider.set_update_policy(gtk.UPDATE_CONTINUOUS)
        self.slider.set_draw_value(self.main.scale_show_value)
        self.slider.set_digits(0)
        self.slider.set_range(0, 100)
        self.slider.set_increments(self.main.scale_increment, 10)
        self.slider.add_events(gtk.gdk.SCROLL_MASK)

        if isinstance(self.slider, gtk.VScale):
            self.align.set_padding(10, 10, 4, 4)
            self.slider.set_size_request(-1, 128)
            self.slider.set_value_pos(gtk.POS_BOTTOM)
            self.slider.set_inverted(True)
        else:
            self.align.set_padding(4, 4, 10, 10)
            self.slider.set_size_request(128, -1)
            self.slider.set_value_pos(gtk.POS_LEFT)

        self.slider.connect("value_changed", self.main.on_volume_changed)
        self.slider.connect("button_press_event", self.on_scale_button_press_event)
        self.slider.connect("button_release_event", self.on_scale_button_release_event)
        self.slider.connect("scroll_event", self.on_scale_scroll_event)

    def on_scale_button_press_event(self, widget, event):
        """ Callback for button_press_event. We want the behaviour you get with the middle button """
        if event.button == 1:
            event.button = 2
        return False

    def on_scale_button_release_event(self, widget, event):
        """ Callback for button_release_event. We want the behaviour you get with the middle button """
        if event.button == 1:
            event.button = 2
        return False

    def on_scale_scroll_event(self, widget, event):
        """ Callback for scroll_event. Forwards event to statusicon """
        self.main.on_scroll_event(widget, event)
        return True

    def on_window_button_press_event(self, widget, event):
        """ Callback for button_press_event. If clicked somewhere else release window """
        if event.type == gtk.gdk.BUTTON_PRESS:
            self.release_grab()
            return True
        return False

    def on_window_key_release_event(self, widget, event):
        """ Callback for key_release_event. On escape key release window """
        if event.keyval == gtk.gdk.keyval_from_name("Escape"):
            self.release_grab()
            return True
        return True

    def on_window_scroll_event(self, widget, event):
        """ Callback for scroll_event. Forwards event to statusicon """
        self.main.on_scroll_event(widget, event)
        return True

    def on_realize(self, widget):
        """ Callback for realize. Move window when realized """
        self.move_window()

    def toggle_window(self):
        """ Toggle scale window visibility """
        if self.win.get_property("visible"):
            self.release_grab()
        else:
            screen, rectangle, orientation = self.main.get_geometry()
            if orientation != self.orientation:
                self.orientation = orientation
                self.init_window()
                self.main.scale = self
                self.main.update()
            if rectangle.x != self.rectangle.x or rectangle.y != self.rectangle.y:
                self.rectangle = rectangle
                self.win.unrealize()
            self.win.show_all()
            self.grab_window()

    def move_window(self):
        """ Move scale window """
        posx, posy = self.get_position()
        self.win.move(posx, posy)

    def grab_window(self):
        """ Grab and focus window """
        self.win.grab_add()
        gtk.gdk.pointer_grab(self.win.window, True,
            gtk.gdk.BUTTON_PRESS_MASK |
            gtk.gdk.BUTTON_RELEASE_MASK |
            gtk.gdk.POINTER_MOTION_MASK |
            gtk.gdk.SCROLL_MASK)

        if gtk.gdk.pointer_is_grabbed():
            if gtk.gdk.keyboard_grab(self.win.window, True) != gtk.gdk.GRAB_SUCCESS:
                self.release_grab()
                return False
        else:
            self.win.grab_remove()
            self.win.hide()
            return False

        self.win.grab_focus()
        return True

    def release_grab(self):
        """ Release grab from window """
        display = self.win.get_display()
        display.keyboard_ungrab()
        display.pointer_ungrab()
        self.win.grab_remove()
        self.win.hide()

    def get_position(self):
        """ Get coordinates to place scale window """
        screen, rectangle, orientation = self.main.get_geometry()
        self.win.set_screen(screen)
        monitor_num = screen.get_monitor_at_point(rectangle.x, rectangle.y)
        monitor = screen.get_monitor_geometry(monitor_num)
        window = self.win.allocation

        if orientation == gtk.ORIENTATION_VERTICAL:
            if monitor.width - rectangle.x == rectangle.width:
                # right panel
                posx = monitor.x + monitor.width - window.width - rectangle.width
            else:
                # left panel
                posx = rectangle.x + rectangle.width
            posy = rectangle.y
        else:
            if (rectangle.y + rectangle.height + window.height <= monitor.y + monitor.height):
                posy = rectangle.y + rectangle.height
            else:
                posy = rectangle.y - window.height

            if (rectangle.x + window.width <= monitor.x + monitor.width):
                posx = rectangle.x
            else:
                posx = monitor.x + monitor.width - window.width

        return posx, posy



import os

def which(prog):
    """ Equivalent of unix which command """
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(prog)
    if fpath:
        if is_exe(prog):
            return prog
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            filename = os.path.join(path, prog)
            if is_exe(filename):
                return filename
    return None



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop

class DBusService(dbus.service.Object):
    """ DBus Service """

    def __init__(self, main_instance):
        """ Constructor """
        loop = DBusGMainLoop()
        self.main = main_instance
        session_bus = dbus.SessionBus(mainloop=loop)
        bus_name = dbus.service.BusName('com.google.code.Volti', bus=session_bus)
        dbus.service.Object.__init__(self, bus_name, '/com/google/code/Volti')

        obj = session_bus.get_object('com.google.code.Volti', '/com/google/code/Volti')
        iface = dbus.Interface(obj, 'com.google.code.Volti')
        iface.connect_to_signal("signal", self.signal_handler)

    @dbus.service.signal('com.google.code.Volti', signature='s')
    def signal(self, signal):
        """ DBus signal """
        pass

    @dbus.service.method('com.google.code.Volti', in_signature='s', out_signature='')
    def emit(self, signal):
        """ DBus method to emit signal """
        self.signal(signal)

    def signal_handler(self, signal):
        """ Handle dbus signals and pass them to main app """
        if signal == 'volume-up':
            self.main.change_volume('up', True)
        elif signal == 'volume-down':
            self.main.change_volume('down', True)
        elif signal == 'mute':
            self.main.change_volume('mute', True)



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import gtk
import pango
from ConfigParser import ConfigParser

PREFS = {
    "card_index": 0,
    "control": "Master",
    "mixer": "alsamixer",
    "run_in_terminal": 1,
    "mixer_internal": 1,
    "mixer_show_values": 1,
    "icon_theme": "Default",
    "scale_increment": 1.0,
    "scale_show_value": 0,
    "show_tooltip": 1,
    "toggle": "mute",
    "keys": 0,
    "keys_backend": "hal",
    "show_notify": 0,
    "notify_timeout": 2.0,
    "notify_position": 0,
    "notify_body": '<span font_desc="14" weight="bold">{volume}</span>\n<small>{card}</small>\n<small>{mixer}</small>'
    }

_PREFERENCES = None

class Preferences:
    """ Preferences window """

    def __init__(self, main_instance):
        """ Constructor """
        self.main = main_instance

        self.cp = ConfigParser()
        self.set_section()

        if not os.path.isfile(self.main.config.config_file):
            self.write_file()

        self.read_file()

    def read_file(self):
        """ Read config file """
        self.cp.read(self.main.config.config_file)
        for option in self.cp.options("global"):
            PREFS[option.lower()] = self.cp.get("global", option).strip()
        PREFS["control"] = self.cp.get(self.section, "control").strip()

    def write_file(self):
        """ Write config file """
        if not os.path.isdir(self.main.config.config_dir):
            try:
                os.makedirs(self.main.config.config_dir)
            except OSError:
                pass
        for section in self.section, "global":
            if not self.cp.has_section(section):
                self.cp.add_section(section)
        for key, val in PREFS.items():
            if key in ["control"]:
                self.cp.set(self.section, key, val)
            else:
                self.cp.set("global", key, val)
        self.cp.write(open(self.main.config.config_file, "w"))

    def open(self, widget=None, data=None):
        """ Open preferences window """
        global _PREFERENCES
        if _PREFERENCES is None:
            _PREFERENCES = Preferences(self.main)
            _PREFERENCES.init_builder()
            _PREFERENCES.init_card_combobox()
            _PREFERENCES.init_treeview()
            _PREFERENCES.init_theme_combobox()
            _PREFERENCES.window.show_all()
        else:
            _PREFERENCES.window.present()

    def close(self, widget=None):
        """ Close preferences window """
        global _PREFERENCES
        start, end = self.notify_body_text.get_buffer().get_bounds()
        body = self.notify_body_text.get_buffer().get_text(start, end)
        PREFS["notify_body"] = body
        self.main.notify_body = body

        self.write_file()
        if _PREFERENCES is not None:
            _PREFERENCES.window.destroy()
            _PREFERENCES = None

    def set_section(self):
        """ Set section name """
        self.section = "card-%s" % PREFS["card_index"]

    def init_builder(self):
        """ Initialize gtk.Builder """
        self.glade = os.path.join(self.main.config.res_dir, "preferences.glade")
        self.tree = gtk.Builder()
        self.tree.set_translation_domain(self.main.config.app_name)
        self.tree.add_from_file(self.glade)

        self.version_label = self.tree.get_object("version_label")
        self.version_label.set_text("%s %s" % (
            self.main.config.app_name.capitalize(), self.main.config.app_version))

        self.window = self.tree.get_object("window")
        self.window.connect("destroy", self.close)
        icon_theme = gtk.icon_theme_get_default()
        if icon_theme.has_icon("multimedia-volume-control"):
            self.window.set_icon_name("multimedia-volume-control")
        else:
            file = os.path.join(
                    self.main.config.res_dir, "icons", "multimedia-volume-control.svg")
            self.window.set_icon_from_file(file)

        self.button_close = self.tree.get_object("button_close")
        self.button_close.connect("clicked", self.close)

        self.button_browse = self.tree.get_object("button_browse")
        self.button_browse.connect("clicked", self.on_browse_button_clicked)

        self.mixer_entry = self.tree.get_object("mixer_entry")
        self.mixer_entry.set_text(PREFS["mixer"])
        self.mixer_entry.connect_after("changed", self.on_entry_changed)

        self.scale_spinbutton = self.tree.get_object("scale_spinbutton")
        self.scale_spinbutton.set_value(float(PREFS["scale_increment"]))
        self.scale_spinbutton.connect("value_changed", self.on_scale_spinbutton_changed)

        self.tooltip_checkbutton = self.tree.get_object("tooltip_checkbutton")
        self.tooltip_checkbutton.set_active(bool(int(PREFS["show_tooltip"])))
        self.tooltip_checkbutton.connect("toggled", self.on_tooltip_toggled)

        self.terminal_checkbutton = self.tree.get_object("terminal_checkbutton")
        self.terminal_checkbutton.set_active(bool(int(PREFS["run_in_terminal"])))
        self.terminal_checkbutton.connect("toggled", self.on_terminal_toggled)

        self.draw_value_checkbutton = self.tree.get_object("draw_value_checkbutton")
        self.draw_value_checkbutton.set_active(bool(int(PREFS["scale_show_value"])))
        self.draw_value_checkbutton.connect("toggled", self.on_draw_value_toggled)

        self.notify_checkbutton = self.tree.get_object("notify_checkbutton")
        self.notify_checkbutton.set_active(bool(int(PREFS["show_notify"])))
        self.notify_checkbutton.connect("toggled", self.on_notify_toggled)

        self.notify_body_text = self.tree.get_object("notify_body_text")
        self.notify_body_text.get_buffer().set_text(PREFS["notify_body"])

        self.position_checkbutton = self.tree.get_object("position_checkbutton")
        self.position_checkbutton.set_active(bool(int(PREFS["notify_position"])))
        self.position_checkbutton.connect("toggled", self.on_position_toggled)

        self.timeout_spinbutton = self.tree.get_object("timeout_spinbutton")
        self.timeout_spinbutton.set_value(float(PREFS["notify_timeout"]))
        self.timeout_spinbutton.connect("value_changed", self.on_timeout_spinbutton_changed)

        self.mixer_internal_checkbutton = self.tree.get_object("mixer_internal_checkbutton")
        self.mixer_internal_checkbutton.set_active(bool(int(PREFS["mixer_internal"])))
        self.mixer_internal_checkbutton.connect("toggled", self.on_mixer_internal_toggled)

        self.mixer_values_checkbutton = self.tree.get_object("mixer_values_checkbutton")
        self.mixer_values_checkbutton.set_active(bool(int(PREFS["mixer_show_values"])))
        self.mixer_values_checkbutton.connect("toggled", self.on_mixer_values_toggled)

        self.keys_checkbutton = self.tree.get_object("keys_checkbutton")
        self.keys_checkbutton.set_active(bool(int(PREFS["keys"])))
        self.keys_checkbutton.connect("toggled", self.on_keys_toggled)

        self.hal_radiobutton = self.tree.get_object("radiobutton_hal")
        self.xlib_radiobutton = self.tree.get_object("radiobutton_xlib")
        if PREFS["keys_backend"] == "hal":
            self.hal_radiobutton.set_active(True)
        elif PREFS["keys_backend"] == "xlib":
            self.xlib_radiobutton.set_active(True)
        self.hal_radiobutton.connect("toggled", self.on_radio_hal_toggled)
        self.xlib_radiobutton.connect("toggled", self.on_radio_xlib_toggled)

        self.mute_radiobutton = self.tree.get_object("radiobutton_mute")
        self.mixer_radiobutton = self.tree.get_object("radiobutton_mixer")
        if PREFS["toggle"] == "mute":
            self.mute_radiobutton.set_active(True)
        elif PREFS["toggle"] == "mixer":
            self.mixer_radiobutton.set_active(True)
        self.mute_radiobutton.connect("toggled", self.on_radio_mute_toggled)
        self.mixer_radiobutton.connect("toggled", self.on_radio_mixer_toggled)

        self.set_keys_sensitive(bool(int(PREFS["keys"])))
        self.set_notify_sensitive(bool(int(PREFS["show_notify"])))
        self.set_mixer_sensitive(bool(int(PREFS["mixer_internal"])))

    def init_card_combobox(self):
        """ Initialize combobox with list of audio cards """
        icon_theme = gtk.icon_theme_get_default()
        if icon_theme.has_icon("audio-card"):
            icon = icon_theme.load_icon(
                    "audio-card", 22, flags=gtk.ICON_LOOKUP_FORCE_SVG)
        else:
            file = os.path.join(
                    self.main.config.res_dir, "icons", "audio-card.svg")
            pixbuf = gtk.gdk.pixbuf_new_from_file(file)
            icon = pixbuf.scale_simple(22, 22, gtk.gdk.INTERP_BILINEAR)

        self.combo_model = gtk.ListStore(int, gtk.gdk.Pixbuf, str)
        for index, card in enumerate(self.main.alsactrl.get_cards()):
            if card is not None:
                self.combo_model.append([index, icon, card])

        card_combobox = self.tree.get_object("card_combobox")
        card_combobox.set_model(self.combo_model)
        card_combobox.set_active(int(PREFS["card_index"]))

        cell1 = gtk.CellRendererPixbuf()
        cell1.set_property("xalign", 0)
        cell1.set_property("xpad", 3)
        card_combobox.pack_start(cell1, False)
        card_combobox.add_attribute(cell1, "pixbuf", 1)

        cell2 = gtk.CellRendererText()
        cell2.set_property("xpad", 10)
        card_combobox.pack_start(cell2, True)
        card_combobox.set_attributes(cell2, text=2)

        card_combobox.connect("changed", self.on_card_combobox_changed)

    def init_treeview(self):
        """ Initialize treeview with mixers """
        self.liststore = gtk.ListStore(bool, str, int)
        for mixer in self.main.alsactrl.get_mixers(int(PREFS["card_index"])):
            active = (mixer == PREFS["control"])
            if active:
                self.liststore.append([active, mixer, pango.WEIGHT_BOLD])
            else:
                self.liststore.append([active, mixer, pango.WEIGHT_NORMAL])

        self.treeview = gtk.TreeView(self.liststore)
        self.treeview.set_headers_visible(False)

        cell1 = gtk.CellRendererToggle()
        cell1.set_radio(True)
        cell1.set_property("activatable", True)
        cell1.connect('toggled', self.on_treeview_toggled, self.liststore)
        column1 = gtk.TreeViewColumn()
        column1.pack_start(cell1, True)
        column1.add_attribute(cell1, 'active', 0)
        self.treeview.append_column(column1)

        cell2 = gtk.CellRendererText()
        column2 = gtk.TreeViewColumn()
        column2.pack_start(cell2, True)
        column2.add_attribute(cell2, 'text', 1)
        column2.add_attribute(cell2, 'weight', 2)
        self.treeview.append_column(column2)

        scrolledwindow = self.tree.get_object("scrolledwindow")
        scrolledwindow.add(self.treeview)

    def init_theme_combobox(self):
        """ Initialize combobox with list of available themes """
        model = gtk.ListStore(int, str)
        theme_combobox = self.tree.get_object("theme_combobox")
        theme_combobox.set_model(model)

        self.themes = self.main.get_icon_themes()
        for index, theme in enumerate(self.themes):
            model.append([index, theme])
            if theme == PREFS["icon_theme"]:
                theme_combobox.set_active(index)

        cell = gtk.CellRendererText()
        theme_combobox.pack_start(cell, True)
        theme_combobox.set_attributes(cell, text=1)

        theme_combobox.connect("changed", self.on_theme_combobox_changed)

    def on_browse_button_clicked(self, widget=None):
        """ Callback for browse_button_clicked event """
        buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                gtk.STOCK_OPEN, gtk.RESPONSE_OK)
        dialog = gtk.FileChooserDialog(
                title=_("Choose external mixer"),
                action=gtk.FILE_CHOOSER_ACTION_OPEN,
                buttons=buttons)
        dialog.set_current_folder("/usr/bin")
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_show_hidden(False)

        file_filter_mixers = gtk.FileFilter()
        file_filter_mixers.set_name(_("Sound Mixers"))
        file_filter_mixers.add_custom(gtk.FILE_FILTER_FILENAME, self.custom_mixer_filter)
        file_filter_all = gtk.FileFilter()
        file_filter_all.set_name(_("All files"))
        file_filter_all.add_pattern("*")
        dialog.add_filter(file_filter_mixers)
        dialog.add_filter(file_filter_all)

        response = dialog.run()
        filename = dialog.get_filename()
        dialog.destroy()

        while gtk.events_pending():
            gtk.main_iteration(False)

        if response == gtk.RESPONSE_OK:
            self.mixer_entry.set_text(filename)
            PREFS["mixer"] = filename
            return filename
        elif response == gtk.RESPONSE_CANCEL:
            return None

    def custom_mixer_filter(self, filter_info=None, data=None):
        """ Custom filter with names of common mixer apps """
        mixers = ["aumix", "alsamixer", "alsamixergui", "gamix",
                "gmixer", "gnome-alsamixer", "gnome-volume-control"]
        if filter_info[2] in mixers:
            return True
        return False

    def on_card_combobox_changed(self, widget=None):
        """ Callback for card_combobox_changed event """
        model = widget.get_model()
        iter = widget.get_active_iter()
        card_index = model.get_value(iter, 0)
        PREFS["card_index"] = card_index

        mixers = self.main.alsactrl.get_mixers(card_index)

        self.set_section()
        if self.cp.has_section(self.section):
            PREFS["control"] = self.cp.get(self.section, "control").strip()
        else:
            PREFS["control"] = mixers[0]

        self.main.update()
        self.liststore.clear()

        for mixer in mixers:
            active = (mixer == PREFS["control"])
            if active:
                self.liststore.append([active, mixer, pango.WEIGHT_BOLD])
            else:
                self.liststore.append([active, mixer, pango.WEIGHT_NORMAL])

    def on_treeview_toggled(self, cell, path, model):
        """ Callback for treeview_toggled event """
        iter = model.get_iter_from_string(path)
        active = model.get_value(iter, 0)
        if not active:
            model.foreach(self.radio_toggle)
            model.set(iter, 0, not active)
            model.set(iter, 2, pango.WEIGHT_BOLD)

            PREFS["control"] = model.get_value(iter, 1)
            self.main.update()
            self.write_file()

    def on_theme_combobox_changed(self, widget=None):
        """ Callback for theme_combobox_changed event """
        model = widget.get_model()
        iter = widget.get_active_iter()
        index = model.get_value(iter, 0)

        icon_theme = self.themes[index]
        PREFS["icon_theme"] = icon_theme
        self.main.icon_theme = icon_theme

        volume = self.main.get_volume()
        icon = self.main.get_icon_name(volume)
        self.main.update_icon(volume, icon)

    def radio_toggle(self, model, path, iter):
        """ Toggles radio buttons status """
        active = model.get(iter, 0)
        if active:
            model.set(iter, 0, not active)
            model.set(iter, 2, pango.WEIGHT_NORMAL)

    def on_scale_spinbutton_changed(self, widget):
        """ Callback for scale_spinbutton_changed event """
        scale_increment = widget.get_value()
        PREFS["scale_increment"] = scale_increment
        self.main.scale_increment = scale_increment

    def on_tooltip_toggled(self, widget):
        """ Callback for tooltip_toggled event """
        active = widget.get_active()
        PREFS["show_tooltip"] = int(active)
        self.main.show_tooltip = active
        if active:
            self.main.update_tooltip(self.main.get_volume())
        else:
            self.main.set_tooltip(None)

    def on_draw_value_toggled(self, widget):
        """ Callback for draw_value_toggled event """
        active = widget.get_active()
        PREFS["scale_show_value"] = int(active)
        self.main.scale.slider.set_draw_value(active)

    def on_terminal_toggled(self, widget):
        """ Callback for terminal_toggled event """
        active = widget.get_active()
        PREFS["run_in_terminal"] = int(active)
        self.main.run_in_terminal = active

    def on_entry_changed(self, widget):
        """ Callback for entry_changed event """
        mixer = widget.get_text()
        PREFS["mixer"] = mixer
        self.main.mixer = mixer

    def on_mixer_internal_toggled(self, widget):
        """ Callback for mixer_internal_toggled event """
        active = widget.get_active()
        PREFS["mixer_internal"] = int(active)
        self.main.mixer_internal = active
        self.set_mixer_sensitive(active)

    def on_mixer_values_toggled(self, widget):
        """ Callback for mixer_values_toggled event """
        active = widget.get_active()
        PREFS["mixer_show_values"] = int(active)

    def on_radio_mute_toggled(self, widget):
        """ Callback for radio_mute_toggled event """
        if widget.get_active():
            PREFS["toggle"] = "mute"
            self.main.toggle = "mute"

    def on_radio_mixer_toggled(self, widget):
        """ Callback for radio_mixer_toggled event """
        if widget.get_active():
            PREFS["toggle"] = "mixer"
            self.main.toggle = "mixer"

    def on_keys_toggled(self, widget):
        """ Callback for keys_toggled event """
        active = widget.get_active()
        PREFS["keys"] = int(active)
        self.main.keys = active
        self.main.init_keys_events()
        self.set_keys_sensitive(active)

    def on_notify_toggled(self, widget):
        """ Callback for notify_toggled event """
        active = widget.get_active()
        PREFS["show_notify"] = int(active)
        self.main.show_notify = active
        self.main.init_notify()
        self.set_notify_sensitive(active)
        if active and self.main.notify:
            volume = self.main.get_volume()
            icon = self.main.get_icon_name(volume)
            self.main.update_notify(volume, icon)

    def on_position_toggled(self, widget):
        """ Callback for position_toggled event """
        active = widget.get_active()
        PREFS["notify_position"] = int(active)
        self.main.notify_position = active
        if self.main.notify:
            self.main.notify.close()
            volume = self.main.get_volume()
            icon = self.main.get_icon_name(volume)
            self.main.update_notify(volume, icon)

    def on_timeout_spinbutton_changed(self, widget):
        """ Callback for spinbutton_changed event """
        timeout = widget.get_value()
        PREFS["notify_timeout"] = timeout
        self.main.notify_timeout = timeout

    def on_radio_hal_toggled(self, widget):
        """ Callback for radio_hal_toggled event """
        if widget.get_active():
            PREFS["keys_backend"] = "hal"
            self.main.keys_backend = "hal"
            self.main.init_keys_events()

    def on_radio_xlib_toggled(self, widget):
        """ Callback for radio_xlib_toggled event """
        if widget.get_active():
            PREFS["keys_backend"] = "xlib"
            self.main.keys_backend = "xlib"
            self.main.init_keys_events()

    def set_mixer_sensitive(self, active):
        """ Set widgets sensitivity """
        self.mixer_values_checkbutton.set_sensitive(active)
        self.mixer_entry.set_sensitive(not active)
        self.button_browse.set_sensitive(not active)
        self.terminal_checkbutton.set_sensitive(not active)

    def set_keys_sensitive(self, active):
        """ Set widgets sensitivity """
        self.hal_radiobutton.set_sensitive(active)
        self.xlib_radiobutton.set_sensitive(active)

    def set_notify_sensitive(self, active):
        """ Set widgets sensitivity """
        self.timeout_spinbutton.set_sensitive(active)
        self.position_checkbutton.set_sensitive(active)
        self.notify_body_text.set_sensitive(active)
        if active and not self.main.notify.check_capabilities():
            self.position_checkbutton.set_sensitive(False)
            self.timeout_spinbutton.set_sensitive(False)



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import alsaaudio as alsa

OLD_VOLUME = 0
MUTED = False

class AlsaControl():
    """ Interface to ALSA mixer API. """

    def __init__(self, prefs):
        """ Constructor """
        try:
            self.prefs = prefs
            self.muted = False
            self.card_index = int(self.prefs["card_index"])
            self.control = self.prefs["control"]
            self.channel = alsa.MIXER_CHANNEL_ALL
            self.mixer = alsa.Mixer(control=self.control, cardindex=self.card_index)

            try:
                assert hasattr(self.mixer, "polldescriptors")
            except AssertionError:
                sys.stderr.write("This program needs pyalsaaudio 0.6 or higher\nExiting\n")
                sys.exit(1)

        except Exception, err:
            sys.stderr.write("%s.%s: can't open %s control for card %s\nerror: %s\n" % (
                __name__, sys._getframe().f_code.co_name, self.control, self.get_card_name(), str(err)))
            try:
                self.mixer = alsa.Mixer(control=self.get_mixers(self.card_index)[0], cardindex=self.card_index)
            except Exception, err:
                sys.stderr.write("%s.%s: can't open first available control for card %s\nerror: %s\nExiting\n" % (
                    __name__, sys._getframe().f_code.co_name, self.get_card_name(), str(err)))
                sys.exit(1)

    def get_descriptors(self):
        """ Returns file descriptors """
        try:
            return self.mixer.polldescriptors()[0]
        except Exception, err:
            sys.stderr.write("%s.%s: %s\n" % (
                __name__, sys._getframe().f_code.co_name, str(err)))
            return None

    def set_volume(self, volume):
        """ Set mixer volume """
        try:
            self.mixer.setvolume(volume, self.channel)
            return True
        except alsa.ALSAAudioError, err:
            sys.stderr.write("%s.%s: %s\n" % (
                __name__, sys._getframe().f_code.co_name, str(err)))
            return False

    def get_volume(self):
        """ Get mixer volume """
        try:
            return self.mixer.getvolume()[0]
        except alsa.ALSAAudioError, err:
            sys.stderr.write("%s.%s: %s\n" % (
                __name__, sys._getframe().f_code.co_name, str(err)))

    def set_mute(self, mute=0):
        """ Set mixer mute status """
        global OLD_VOLUME, MUTED
        try:
            self.mixer.setmute(mute, self.channel)
        except alsa.ALSAAudioError:
            # mixer doesn't have mute switch
            if mute == 1:
                OLD_VOLUME = self.get_volume()
                self.set_volume(0)
                MUTED = True
            else:
                self.set_volume(OLD_VOLUME)
                MUTED = False

    def is_muted(self):
        """ Returns mixer mute status """
        global OLD_VOLUME, MUTED
        try:
            if self.mixer.getmute()[0] == 1:
                return True
        except alsa.ALSAAudioError:
            if MUTED:
                return True
        return False

    def get_card_name(self):
        """ Returns card name """
        try:
            return alsa.cards()[self.card_index]
        except IndexError, err:
            sys.stderr.write("%s.%s: %s\n" % (
                __name__, sys._getframe().f_code.co_name, str(err)))

    def get_mixer_name(self):
        """ Returns mixer name """
        try:
            return self.mixer.mixer()
        except alsa.ALSAAudioError, err:
            sys.stderr.write("%s.%s: %s\n" % (
                __name__, sys._getframe().f_code.co_name, str(err)))

    def get_cards(self):
        """ Returns cards list """
        cards = []
        acards = alsa.cards()
        for index, card in enumerate(acards):
            try:
                self.get_mixers(index)[0]
            except IndexError:
                cards.append(None)
            else:
                cards.append(acards[index])
        return cards

    def get_mixers(self, card_index=0):
        """ Returns mixers list """
        mixers = []
        for mixer in alsa.mixers(card_index):
            try:
                m = alsa.Mixer(control=mixer, cardindex=card_index)
                if 'Playback Volume' in m.volumecap():
                    mixers.append(mixer)
            except alsa.ALSAAudioError:
                pass
        return mixers



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import threading

import gtk
import gobject
from Xlib.display import Display
from Xlib import X, XK


class XlibEvent(gobject.GObject, threading.Thread):
    """ Handle multimedia keys via Xlib """

    __gsignals__ = {
            'volume-up': (gobject.SIGNAL_RUN_FIRST, None, ()),
            'volume-down': (gobject.SIGNAL_RUN_FIRST, None, ()),
            'mute': (gobject.SIGNAL_RUN_FIRST, None, ())
        }

    def __init__(self, main_instance):
        """ Constructor """
        gobject.GObject.__init__(self)
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.running = False
        self.main = main_instance

        try:
            XK.load_keysym_group("xf86")
        except ImportError, err:
            sys.stderr.write("Xlib backend needs python-xlib 0.15rc1 or higher\n")
            raise ImportError(str(err))

        self.display = Display()
        self.screen = self.display.screen()
        self.root = self.screen.root
        self.root.change_attributes(event_mask = X.KeyPressMask)

        self.up_keys = self.get_keycodes(self.display.keysym_to_keycodes(XK.XK_XF86_AudioRaiseVolume))
        self.down_keys = self.get_keycodes(self.display.keysym_to_keycodes(XK.XK_XF86_AudioLowerVolume))
        self.mute_keys = self.get_keycodes(self.display.keysym_to_keycodes(XK.XK_XF86_AudioMute))
        self.keycodes = [self.up_keys, self.down_keys, self.mute_keys]

        self.grab()

        self.connect("volume-up", self.button_handler, "volume-up")
        self.connect("volume-down", self.button_handler, "volume-down")
        self.connect("mute", self.button_handler, "mute")

    def get_keycodes(self, keys):
        """ Returns keycodes without modifiers """
        keycodes = []
        for keycode, index in keys:
            if keycode not in keycodes:
                keycodes.append(keycode)
        return keycodes

    def grab(self):
        """ Grab keys, will print error if some other app already have those keys grabbed """
        for keys in self.keycodes:
            for keycode in keys:
                self.root.grab_key(keycode, X.AnyModifier, True, X.GrabModeAsync, X.GrabModeAsync)

    def ungrab(self):
        """ Ungrab keys """
        for keys in self.keycodes:
            for keycode in keys:
                self.root.ungrab_key(keycode, X.AnyModifier, self.root)

    def button_handler(self, sender, event):
        """ Handle button events and pass them to main app """
        if event == 'volume-up':
            self.main.change_volume('up', True)
        elif event == 'volume-down':
            self.main.change_volume('down', True)
        elif event == 'mute':
            self.main.change_volume('mute', True)

    def signal(self, signal):
        """ Emit signal """
        gtk.gdk.threads_enter()
        self.emit(signal)
        gtk.gdk.threads_leave()
        return False

    def run(self):
        """ Start thread """
        self.running = True
        while self.running:
            event = self.display.next_event()
            if event.type == X.KeyPress:
                if event.detail in self.up_keys:
                    gobject.idle_add(self.signal, "volume-up")
                elif event.detail in self.down_keys:
                    gobject.idle_add(self.signal, "volume-down")
                elif event.detail in self.mute_keys:
                    gobject.idle_add(self.signal, "mute")

    def stop(self):
        """ Stop thread """
        self.running = False
        self.ungrab()
        self.display.close()



#!/usr/bin/env python

import os
import sys
import glob
import subprocess
from distutils.core import setup
from distutils.command.install_data import install_data
from distutils.command.build import build
from distutils.dep_util import newer
from distutils.log import info

from src.config import Config
config = Config()

PO_DIR = 'po'
MO_DIR = os.path.join('build', 'mo')
ICON_DIR = os.path.join("data", "icons")

class BuildLocales(build):
  def run(self):
    build.run(self)

    for po in glob.glob(os.path.join(PO_DIR, '*.po')):
      lang = os.path.basename(po[:-3])
      mo = os.path.join(MO_DIR, lang, config.app_name + '.mo')

      directory = os.path.dirname(mo)
      if not os.path.exists(directory):
        os.makedirs(directory)

      if newer(po, mo):
        info('compiling %s -> %s' % (po, mo))
        try:
          rc = subprocess.call(['msgfmt', po, '-o', mo])
          if rc != 0:
            raise Warning, "msgfmt returned %d" % rc
        except Exception, e:
          print "Building gettext files failed."
          print "%s: %s" % (type(e), str(e))
          sys.exit(1)

class Install(install_data):
  def run(self):
    self.data_files.extend(self._find_mo_files())
    self.data_files.extend(self._find_icons())
    install_data.run(self)

  def _find_mo_files(self):
    data_files = []
    for mo in glob.glob(os.path.join(MO_DIR, '*', config.app_name + '.mo')):
        lang = os.path.basename(os.path.dirname(mo))
        dest = os.path.join('share', 'locale', lang, 'LC_MESSAGES')
        data_files.append((dest, [mo]))
    return data_files

  def _find_icons(self):
    data_files = []
    for filepath in os.listdir(ICON_DIR):
        filepath_full = os.path.join(ICON_DIR, filepath)
        if os.path.isdir(filepath_full):
            if not filepath.startswith("."):
                for size in os.listdir(filepath_full):
                    size_full = os.path.join(ICON_DIR, filepath, size)
                    if os.path.isdir(size_full) and not size.startswith("."):
                        icons = glob.glob(os.path.join(
                            ICON_DIR, filepath, size, "*"))
                        targetpath = os.path.join(
                                "share", "volti", "icons", filepath, size)
                        data_files.append((targetpath, icons))
        else:
            sourcepath = os.path.join(
                    ICON_DIR, filepath)
            targetpath = os.path.join(
                    "share", "volti", "icons")
            data_files.append((targetpath, [sourcepath]))
    return data_files

setup(name = config.app_name,
        version = config.app_version,
        description = "GTK+ application for controlling audio volume from system tray/notification area",
        author = "Milan Nikolic",
        author_email = "gen2brain@gmail.com",
        license = "GNU GPLv3",
        url = "http://code.google.com/p/volti/",
        download_url = "http://volti.googlecode.com/files/%s-%s.tar.gz " % (config.app_name, config.app_version),
        packages = ["volti"],
        package_dir = {"volti": "src"},
        scripts = ["volti", "volti-mixer", "volti-remote"],
        requires = ["gtk", "gobject", "cairo", "alsaaudio", "dbus", "Xlib"],
        platforms = ["Linux"],
        cmdclass = {'build': BuildLocales, 'install_data': Install},
        data_files = [("share/volti", ["data/preferences.glade"]),
                    ("share/applications", ["data/volti.desktop"]),
                    ("share/man/man1", ["doc/volti.1", "doc/volti-remote.1", "doc/volti-mixer.1"])]
        )



# -*- coding: utf-8 -*-

# Author: Jon Nettleton <jon.nettleton@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

import dbus
from dbus.mainloop.glib import DBusGMainLoop

class DBusEvent:
    """ Handle multimedia keys via dbus/hal """

    def __init__(self, main_instance):
        """ Constructor """
        loop = DBusGMainLoop()
        self.main = main_instance
        bus = dbus.SystemBus(mainloop=loop)
        for udi in self.get_inputs():
            obj = bus.get_object('org.freedesktop.Hal', udi)
            iface = dbus.Interface(obj, 'org.freedesktop.Hal.Device')
            iface.connect_to_signal("Condition", self.button_handler, path_keyword="path")

    def hal_manager(self):
        """ Hal manager """
        bus = dbus.SystemBus()
        obj = bus.get_object('org.freedesktop.Hal', '/org/freedesktop/Hal/Manager')
        return dbus.Interface(obj, 'org.freedesktop.Hal.Manager')

    def get_inputs(self):
        """ Get keys """
        return self.hal_manager().FindDeviceByCapability("input.keys")

    def button_handler(self, sender, destination, path):
        """ Handle button events and pass them to main app """
        if sender == 'ButtonPressed':
            if destination == 'volume-up':
                self.main.change_volume('up', True)
            elif destination == 'volume-down':
                self.main.change_volume('down', True)
            elif destination.startswith('mute'):
                self.main.change_volume('mute', True)



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

class Config:

    def __init__(self):
        """ Constructor """
        self.app_name = "volti"
        self.app_version = "0.2.2"
        self.res_dir = None
        self.locale_dir = None

        if not os.path.isdir(os.path.join(".","src")):
            for base in ["/usr/share", "/usr/local/share"]:
                if os.path.isdir(os.path.join(base, self.app_name)):
                    self.res_dir = os.path.join(base, self.app_name)
                    self.locale_dir = os.path.join(base, "locale")
                    break

        if not self.res_dir:
            self.res_dir = "data"

        self.config_dir = os.path.expanduser(os.path.join("~", ".config", self.app_name))
        self.config_file = os.path.join(self.config_dir, "config")



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
# Based on code from http://code.google.com/p/rox-volume
# by Kenneth Hayber <ken@hayber.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import gettext
from ConfigParser import ConfigParser

import gtk
import gobject
import alsaaudio as alsa

from config import Config
CONFIG = Config()

CHANNEL_LEFT  = 0
CHANNEL_RIGHT = 1
CHANNEL_MONO  = 2

_STEREO	= 1
_LOCK   = 2
_REC    = 4
_MUTE   = 8

class Mixer(gtk.Window):
    """ Volti Mixer Application"""
    def __init__(self):
        gtk.Window.__init__(self)

        self.cp = ConfigParser()
        self.cp.read(CONFIG.config_file)

        self.lock_mask = {}
        self.alsa_channels = {}

        self.set_title("Volti Mixer")
        self.set_resizable(True)
        icon_theme = gtk.icon_theme_get_default()
        if icon_theme.has_icon("multimedia-volume-control"):
            self.set_icon_name("multimedia-volume-control")
        else:
            file = os.path.join(
                    CONFIG.res_dir, "icons", "multimedia-volume-control.svg")
            self.set_icon_from_file(file)

        notebook = gtk.Notebook()
        notebook.set_tab_pos(gtk.POS_TOP)
        self.add(notebook)

        try:
            show_values = bool(int(self.cp.get("global", "mixer_show_values")))
        except:
            show_values = False

        for card_index, card_name in enumerate(alsa.cards()):
            vbox = gtk.VBox()
            hbox = gtk.HBox(False, 10)
            frame = gtk.Frame()
            label = gtk.Label(card_name)

            align = gtk.Alignment(xscale=1, yscale=1)
            align.set_padding(10, 10, 10, 10)

            align.add(hbox)
            vbox.add(align)
            frame.add(vbox)
            notebook.append_page(frame, label)

            try:
                self.lock_mask[card_index] = int(
                        self.cp.get("card-%d" % card_index, "mask_lock"))
            except:
                self.lock_mask[card_index] = 0

            n = 0
            self.get_channels(card_index)
            for channel, id in self.alsa_channels[card_index]:
                mixer = alsa.Mixer(channel, id, card_index)

                option_mask = option_value = 0

                if not len(mixer.volumecap()):
                    continue

                if len(mixer.getvolume()) > 1:
                    option_mask |= _STEREO
                    option_mask |= _LOCK

                if self.lock_mask[card_index] & (1 << n):
                    option_value |= _LOCK

                try:
                    if mixer.getrec():
                        option_mask |= _REC
                    if mixer.getrec()[0]:
                        option_value |= _REC
                except:
                    pass

                try:
                    if mixer.getmute():
                        option_mask |= _MUTE
                    if mixer.getmute()[0]:
                        option_value |= _MUTE
                except:
                    pass

                volume = MixerControl(n, option_mask, option_value,
                        show_values, card_index, channel)
                volume.set_level(self.get_volume(n, card_index))
                volume.connect("volume_changed", self.adjust_volume)
                volume.connect("volume_setting_toggled", self.setting_toggled)
                hbox.pack_start(volume, True, True)
                n += 1

        self.add_events(gtk.gdk.BUTTON_PRESS_MASK)
        self.connect('delete_event', self.quit)
        self.show_all()

    def get_channels(self, card_index):
        try:
            self.alsa_channels[card_index] = []
            for channel in alsa.mixers(card_index):
                id = 0
                while (channel, id) in self.alsa_channels[card_index]:
                    id += 1
                mixer = alsa.Mixer(channel, id, card_index)
                if len(mixer.volumecap()):
                    self.alsa_channels[card_index].append((channel, id))
        except:
            pass

    def setting_toggled(self, vol, channel, button, val, card_index):
        """ Handle checkbox toggles """
        ch, id = self.alsa_channels[card_index][channel]
        mixer = alsa.Mixer(ch, id, card_index)

        if button == _MUTE:
            mixer.setmute(val)

        if button == _LOCK:
            if val:
                self.lock_mask[card_index] |= (1<<channel)
            else:
                self.lock_mask[card_index] &= ~(1<<channel)

        if button == _REC:
            mixer.setrec(val)

    def adjust_volume(self, vol, channel, volume1, volume2, card_index):
        """ Track changes to the volume controls """
        self.set_volume((volume1, volume2), channel, card_index)

    def set_volume(self, volume, channel, card_index):
        """ Set the playback volume """
        ch, id = self.alsa_channels[card_index][channel]
        mixer = alsa.Mixer(ch, id, card_index)
        try:
            mixer.setvolume(volume[0], 0)
            mixer.setvolume(volume[1], 1)
        except:
            pass

    def get_volume(self, channel, card_index):
        """ Get the current sound card setting for specified channel """
        ch, id = self.alsa_channels[card_index][channel]
        mixer = alsa.Mixer(ch, id, card_index)
        vol = mixer.getvolume()
        if len(vol) == 1:
            return (vol[0], vol[0])
        return (vol[0], vol[1])

    def quit(self, element=None, event=None):
        """ Exit main loop """
        for card_index, card_name in enumerate(alsa.cards()):
            section = "card-%d" % card_index
            if not self.cp.has_section(section):
                self.cp.add_section(section)
            self.cp.set(section, "mask_lock", self.lock_mask[card_index])
        self.cp.write(open(CONFIG.config_file, "w"))
        gtk.main_quit()


class MixerControl(gtk.Frame):
    """
    A Class that implements a volume control (stereo or mono) for a sound card
    mixer. Each instance represents one mixer channel on the sound card.
    """

    __gsignals__ = {
            'volume_changed': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_BOOLEAN,
                (gobject.TYPE_INT, gobject.TYPE_INT, gobject.TYPE_INT, gobject.TYPE_INT)),
            'volume_setting_toggled': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_BOOLEAN,
                (gobject.TYPE_INT, gobject.TYPE_INT, gobject.TYPE_INT, gobject.TYPE_INT))
        }

    def __init__(self, channel, option_mask, option_value, show_value, card_index, label=None):
        """
        Create a volume control widget
        'channel' specifies the audio device mixer channel.
        'option_mask' configures the widget while 'option_value' sets the actual
        value of the corresponding mask (e.g. 'option_mask |= _MUTE' shows the mute
        checkbox while 'option_value |= _MUTE' causes it to be checked by default)
        'show_value' controls whether the volume text is displayed or not.
        'card_index' is the index of the sound card
        'label' is the name of the channel (e.g. 'PCM).

        The widget supports two signals 'volume_changed' and 'volume_setting_toggled'.
        'volume_changed' always sends left and right volume settings regardless of
        whether the control is locked or mono.

        'volume_setting_toggled' notifies the parent of changes in the optional checkboxes.
        """
        gtk.Frame.__init__(self, label)

        self.card_index = card_index
        self.rec = self.lock = self.stereo = self.mute = False
        if option_mask & _LOCK:
            self.lock = True
            self.channel_locked = option_value & _LOCK

        if option_mask & _REC:
            self.rec = True
            self.channel_rec = option_value & _REC

        if option_mask & _MUTE:
            self.mute = True
            self.channel_muted = option_value & _MUTE

        if option_mask & _STEREO:
            self.stereo = True

        self.channel = channel
        self.vol_left = self.vol_right = 0

        vbox = gtk.VBox()
        control_box = gtk.HBox(True, 0)
        option_box = gtk.HBox(True, 0)

        align = gtk.Alignment(xscale=1, yscale=1)
        align.set_padding(5, 5, 5, 5)
        self.set_label_align(0.5, 0.8)

        align.add(vbox)
        self.add(align)

        self.volume1 = gtk.Adjustment(0.0, 0.0, 100.0, 1.0, 10.0, 0.0)
        if self.stereo:
            self.volume1_handler_id = self.volume1.connect('value_changed', self.value_changed,
                    channel, CHANNEL_LEFT, card_index)
        else:
            self.volume1_handler_id = self.volume1.connect('value_changed', self.value_changed,
                    channel, CHANNEL_MONO, card_index)

        volume1_control = gtk.VScale(self.volume1)
        volume1_control.set_inverted(True)
        volume1_control.set_draw_value(show_value)
        volume1_control.set_digits(0)
        volume1_control.set_size_request(-1, 250)
        volume1_control.set_value_pos(gtk.POS_TOP)
        control_box.pack_start(volume1_control, True, True)

        if self.stereo:
            self.volume2 = gtk.Adjustment(0.0, 0.0, 100.0, 1.0, 10.0, 0.0)
            self.volume2_handler_id = self.volume2.connect('value_changed', self.value_changed,
                    channel, CHANNEL_RIGHT, card_index)

            volume2_control = gtk.VScale(self.volume2)
            volume2_control.set_inverted(True)
            volume2_control.set_draw_value(show_value)
            volume2_control.set_digits(0)
            volume2_control.set_size_request(-1, 300)
            volume2_control.set_value_pos(gtk.POS_TOP)
            control_box.pack_start(volume2_control, True, True)

        if self.rec:
            self.rec_element = self.toggle_element(
                    self.channel_rec, channel, _REC)
            option_box.pack_start(self.rec_element, False, False)

        if self.mute:
            mute_element = self.toggle_element(
                    self.channel_muted, channel, _MUTE)
            option_box.pack_start(mute_element, False, False)

        if self.stereo and self.lock:
            lock_element = self.toggle_element(
                    self.channel_locked, channel, _LOCK)
            option_box.pack_start(lock_element, False, False)

        self.control1 = volume1_control
        if self.stereo:
            self.control2 = volume2_control

        vbox.pack_start(control_box, True, True)
        vbox.pack_start(option_box, False, False, 5)

        self.show_all()

    def set_level(self, level):
        """
        Allow the volume settings to be passed in from the parent.
        'level' is a tuple of integers from 0-100 as (left, right).
        """
        self.volume1.handler_block(self.volume1_handler_id)
        self.volume1.set_value(level[0])
        self.volume1.handler_unblock(self.volume1_handler_id)
        if self.stereo:
            self.volume2.handler_block(self.volume2_handler_id)
            self.volume2.set_value(level[1])
            self.volume2.handler_unblock(self.volume2_handler_id)

    def get_level(self):
        """
        Return the current widget's volume settings as a tuple of
        integers from 0-100 as (left, right)
        """
        return (self.vol_left, self.vol_right)

    def value_changed(self, vol, channel, channel_lr, card_index):
        """
        Track changes in the volume controls and pass them back to the parent
        via the 'volume_changed' signal.
        """
        if channel_lr == CHANNEL_LEFT:
            self.vol_left = int(vol.get_value())
            if self.lock and self.channel_locked:
                self.vol_right = self.vol_left
                self.volume2.handler_block(self.volume2_handler_id)
                self.volume2.set_value(vol.get_value())
                self.volume2.handler_unblock(self.volume2_handler_id)

        elif channel_lr == CHANNEL_RIGHT:
            self.vol_right = int(vol.get_value())
            if self.lock and self.channel_locked:
                self.vol_left = self.vol_right
                self.volume1.handler_block(self.volume1_handler_id)
                self.volume1.set_value(vol.get_value())
                self.volume1.handler_unblock(self.volume1_handler_id)
        else:
            self.vol_left = self.vol_right = int(vol.get_value())
        self.emit("volume_changed", channel, self.vol_left, self.vol_right, card_index)

    def check(self, button, channel, id):
        """
        Process the various checkboxes/buttons and signal the parent when they change
        via the 'volume_setting_toggled' signal.
        """
        active = button.get_active()
        if id == _LOCK:
            self.channel_locked = not self.channel_locked
            if self.channel_locked:
                avg_vol = (self.vol_left+self.vol_right)/2
                self.volume1.set_value(avg_vol)
                self.volume2.set_value(avg_vol)
        elif id == _MUTE:
            self.channel_muted = not self.channel_muted
        elif id == _REC:
            self.channel_rec = not self.channel_rec
        button.set_property("image", self.button_image(id, active))
        self.emit('volume_setting_toggled', channel, id, button.get_active(), self.card_index)

    def toggle_element(self, active, channel, id):
        button = gtk.ToggleButton()
        image = self.button_image(id, active)
        button.set_property("image", image)
        button.set_active(active)
        button.connect('toggled', self.check, channel, id)
        return button

    def button_image(self, id, active):
        if id == _LOCK:
            icon = "mixer-lock.png" if active else "mixer-no-lock.png"
        elif id == _MUTE:
            icon = "mixer-muted.png" if active else "mixer-no-muted.png"
        elif id == _REC:
            icon = "mixer-record.png" if active else "mixer-no-record.png"
        image = gtk.Image()
        image.set_from_file(os.path.join(CONFIG.res_dir, "icons", icon))
        image.show()
        return image

    def show_values(self, show_value):
        self.control1.set_draw_value(show_value)
        if self.stereo:
            self.control2.set_draw_value(show_value)



# -*- coding: utf-8 -*-

# Author: Milan Nikolic <gen2brain@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import dbus

class Notification:
    """ Desktop notifications """

    def __init__(self, main_instance):
        """ Constructor """
        self.main = main_instance
        bus = dbus.SessionBus()
        obj = bus.get_object('org.freedesktop.Notifications', '/org/freedesktop/Notifications')
        self.notify = dbus.Interface(obj, 'org.freedesktop.Notifications')
        self.server_capable = self.check_capabilities()
        self.title = '' if self.server_capable else 'volume'
        self.last_id = dbus.UInt32()

    def check_capabilities(self):
        info = self.notify.GetServerInformation()
        if info[0] == "notify-osd":
            return False
        return True

    def show(self, icon, message, duration, volume):
        """ Show the notification """
        body = self.format(message, volume)
        hints = {"urgency": dbus.Byte(0), "desktop-entry": dbus.String("volti")}
        if self.main.notify_position and self.server_capable:
            hints["x"], hints["y"] = self.get_position()
        self.last_id = self.notify.Notify('volume', self.last_id, icon, self.title, body, [], hints, duration * 1000)

    def close(self):
        """ Close the notification """
        if self.last_id:
            self.notify.CloseNotification(self.last_id)

    def get_position(self):
        """ Returns status icon center coordinates """
        screen, rectangle, orientation = self.main.get_geometry()
        posx = rectangle.x + rectangle.width/2
        posy = rectangle.y + rectangle.height/2
        return posx, posy

    def format(self, message, volume):
        """ Format notification body """
        var, card_name, mixer_name = self.main.get_status_info(volume)
        message = message.replace('{volume}', '%s%s' % (volume, var))
        message = message.replace('{card}', '%s: %s' % (_("Card"), card_name))
        message = message.replace('{mixer}', '%s: %s' % (_("Mixer"), mixer_name))
        return message
